#import <QuartzCore/CIImageAccumulator.h>
#import "StructureRecordAdv.h"
#import "QCGLCGImage.h"
#import "QCGLCIImage.h"
#import "QCGLBitmapImage.h"
#import "StructureRecordAdvUI.h"

@implementation StructureRecordAdv : QCPatch

@class QCGLCVImage;

+ (int)executionMode
{
	return 2;
}
+ (BOOL)allowsSubpatches
{
	return NO;
}
+ (int)timeMode
{
	return 1;
}

+ (Class)inspectorClassWithIdentifier:(id)fp8
{
        return [StructureRecordAdvUI class];
}

- (id)initWithIdentifier:(id)fp8
{
	id z=[super initWithIdentifier:fp8];

	if(z)
	{
		/* inputMode indexValues: (a bitmap)
			Bit three (&8): nil/BAO
			Bit two (&4): Level/Signal
			Bit one (&2): Stop/Loop
			Bit zero (&1): Front/Back

			0000 nlsf
			0001 nlsb
			0010 nllf
			0011 nllb

			0100 nssf
			0101 nssb
			0110 nslf
			0111 nslb

			1000 blsf
			1001 blsb
			1010 bllf
			1011 bllb

			1100 bssf -
			1101 bssb -
			1110 bslf -
			1111 bslb -
		*/
		[inputMode setMaxIndexValue:11];
		[inputMaxSamples setIndexValue:10];	// don't start in unlimited mode by default :)
		index = 0;
		buffer = nil;
		triggered = FALSE;
	}

	return z;
}

- (NSDictionary*)attributes
{
	NSMutableDictionary *attr = [[super attributes] mutableCopy];
	[attr setObject:@"Kineme Structure Record" forKey:@"name"];
	return attr;
}

- (id)setup:(QCOpenGLContext *)context
{
	return context;
}

- (void)cleanup:(QCOpenGLContext *)context
{
	if(buffer != nil)
	{
		[buffer removeAllObjects];
		[buffer release];
		buffer = nil;
	}
}


- (void)enable:(QCOpenGLContext *)context
{


}
- (void)disable:(QCOpenGLContext *)context
{


}

/* this helper function clones objects that don't adhere to the NSCopying protocol correctly, intelligently, or at all. 
	Currently, we handle: 
	   * QCGLImages - copy is way too expensive
	     * Video Input returns QCGLBitmapImage (Not stable for iSight...)
		 * Image Importer returns QCGLImage
		 * Image with Movie returns QCGLCVImage
		 * stuff passed through CIFilters returns QCGLCIImage
	   * QCStructure (NSCopying doesn't work on this class)
 */
id copyItem(id object, BOOL safeMode)
{
	id newObject;
	if( [object isKindOfClass:[QCGLImage class]] )
	{
		// special case: QCGLCGImages are super-expensive to copy/accumulate, so we just copy the pointer and retain it.
		// this will be wrong if something animates CGImages, but I've not seen any patches do that...
		if( [object class] == [QCGLCGImage class] )
		{
			//NSLog(@"Copying QCGLCGImage");
			newObject = object;//[[QCGLCGImage alloc] initWithCGImage:[object CGImage] options:nil];
			[newObject retain];
		}
		else if(safeMode) // aka Safe Mode
		{
			/* This mode works by copying raw CGImages.  slow, lots of ram usage, but stable vs. iSight */
			//NSLog(@"Recording in Safe Mode");
			newObject = [CIImage imageWithCGImage:[object CGImage]];
			[newObject retain];
		}
		else	// non-safe Mode
		{
			CGRect r;
			r.origin.x=0.0;
			r.origin.y=0.0;
			r.size.width=[object pixelsWide];
			r.size.height=[object pixelsHigh];
		
			CIImageAccumulator *cia = [CIImageAccumulator imageAccumulatorWithExtent:r format:kCIFormatARGB8];
			[cia setImage:[object CIImage]];
			newObject = cia;
			[newObject retain];
		}
	}
	
	else if ([object class] == [QCStructure class])
	{
		// This is kind of a hack.  Structures store their key/value pairs in 2 GFLists.  We simply reuse these lists
		// to perform a "shallow" copy (meaning we only clone pointers, not data).  
		GFList *list = [object _list];
		QCStructure *qcs = [[QCStructure alloc] 
			initWithMembers:(id *)[list _values]
			forKeys:(id *)[list _keys]
			count:[object count]];
		
		newObject = qcs;
	}
	else
	{
		// These next 2 lines are helpful when extending the blender, so we know what kind of object we're pulling out of ports
		// Watch out for Private Inner Classes though (e.g. NSCalibratedRGBColor instead of NSColor)
		//NSLog(@"dumb copy for object class: %@ [%@]",[object class],object);
		//NSLog(@"object superclass: %@",[object superclass]);
		newObject = [object copy];
	}
		
	return newObject;
}

/* convenience functions */
void addToFront(NSMutableArray *buffer, id object, BOOL safeMode)
{
	id newItem = copyItem(object,safeMode);
	[buffer insertObject: newItem atIndex:0];
	[newItem release];
}

void addToBack(NSMutableArray *buffer, id object, int index, BOOL safeMode)
{
	id newItem = copyItem(object,safeMode);
	// can't add past the end of the array
	if(index >= [buffer count])
		[buffer addObject:newItem];
	else // replace mode
		[buffer replaceObjectAtIndex:index withObject:newItem];
	[newItem release];
}

/* -- Horribly complicated inputMode behaviour table
0000 non-Boa, level, stop on full, front
0001 non-Boa, level, stop on full, back
   * These two modes record whenever gate is high, and stop when recorded is inputMaxSamples

0010 non-Boa, level, loop on full, front
0011 non-Boa, level, loop on full, back
   * These two modes record whenever gate is high

0100 non-Boa, signal, stop on full, front
0101 non-Boa, signal, stop on full, back
   * These two modes record when gate signals, and stop when recorded is inputMaxSamples
   
0110 non-Boa, signal, loop on full, front
0111 non-Boa, signal, loop on full, back
   * These two modes record when gate signals

1000 Boa, level, stop on full, front
1001 Boa, level, stop on full, back
   * These two modes record when gate signals, and stop when recorded is inputMaxSamples

1010 Boa, level, loop on full, front
1011 Boa, level, loop on full, back
   * These two modes record when gate signals, and loop when gate signals

recorded is only used in stop modes, and is reset by clear, or gate going low after the buffer is full.
*/
- (BOOL)execute:(QCOpenGLContext *)context time:(double)time arguments:(NSDictionary *)arguments
{
	int mode;
	BOOL	signaled = FALSE;
	
	// can't use nil inputs
	if([inputVirtual rawValue] == nil)
		return YES;
		
	mode = [inputMode indexValue];

	// clear on clear or on gate rising edge if clear-on-gate is set
	if([inputClear booleanValue] == TRUE || [inputClearWithObject booleanValue] == TRUE ||
		([inputClearOnGate booleanValue] == TRUE && oldGateValue == FALSE && [inputGate booleanValue] == TRUE))
	{
		index = 0;
		recorded = 0;
		triggered = FALSE;
		if(buffer != nil)
		{	/* empty the buffer */
			[buffer release];
			buffer = nil;
		}
		if([inputClearWithObject booleanValue] == TRUE)
		{
			buffer = [[NSMutableArray alloc] initWithCapacity:1+[inputMaxSamples indexValue]];
			id newItem = copyItem([inputVirtual rawValue],cgMode);
			while([buffer count] < [inputMaxSamples indexValue])
				[buffer addObject:newItem];
			[newItem release];
		}
		else
			[outputStruct setStructureValue:nil];
	}
	
	if([inputGate booleanValue] == TRUE)
	{
		if(mode & kStructureRecordBAO && recorded < [inputMaxSamples indexValue])
			triggered = TRUE;
		if(oldGateValue == FALSE && mode & kStructureRecordSignal)
			signaled = TRUE;
	}
	
	/* record block proper */
	if(	(mode < 4 && [inputGate booleanValue] == TRUE) ||		/* record on gate high modes ...*/
		(mode >=4 && mode < 8 && signaled) ||					/* ... signaled modes ... */
		(mode >=8 && triggered ))								/* ... buffer at once modes */
	{
		if(buffer == nil)
			buffer = [[NSMutableArray alloc] initWithCapacity:1+[inputMaxSamples indexValue]];
			
		if( (mode & kStructureRecordLoop) ||					/* record loop mode ...*/
			(recorded < [inputMaxSamples indexValue]))			/* ... stop on full mode */
		{
			++recorded;
			if(mode & kStructureRecordBack)
			{
				addToBack(buffer, [inputVirtual rawValue], index, cgMode);
				[outputIndex setIndexValue:index];
				if( [inputMaxSamples indexValue] )
					index = (index+1) % [inputMaxSamples indexValue];
				else
					index++;
			}
			else
			{
				addToFront(buffer, [inputVirtual rawValue], cgMode);
				[outputIndex setIndexValue:0];
			}
		}
		
		/* chop off any stragglers from the end [useful when user resizes buffer] */
		while([buffer count] > [inputMaxSamples indexValue])
			[buffer removeLastObject];
		/* postprocess the array to extract usable types */
		NSMutableArray *processed = [[NSMutableArray alloc] initWithCapacity:[buffer count]];// initWithArray:buffer];
		unsigned int i;
		for(i=0;i<[buffer count];++i)
			if( [[buffer objectAtIndex:i] class] == [CIImageAccumulator class] )
			{
				QCGLCIImage *img = [[QCGLCIImage alloc] initWithCIImage:[[buffer objectAtIndex:i] image] options:nil];
				[processed addObject:img];
				[img release];
				//[processed replaceObjectAtIndex:i withObject:[[processed objectAtIndex:i] image]];
				/*QCGLCIImage *img = [[QCGLCIImage alloc] initWithCIImage:[[processed objectAtIndex:i] image] options:nil];
				[processed replaceObjectAtIndex:i 
					withObject:img];
				[img release];*/
			}
			else
				[processed addObject:[buffer objectAtIndex:i]];
                        
		structure = [[QCStructure alloc] initWithArray:processed];
		[processed release];
		//structure = [[QCStructure alloc] initWithArray:buffer];
		
		[outputStruct setStructureValue: structure];
		[structure release];	
	}
	
	/* if we're in BAO loop mode, we reset recorded so we can continue when we're full */
	if(mode & kStructureRecordBAO && mode & kStructureRecordLoop && recorded >= [inputMaxSamples indexValue])
		recorded = 0;

	
	oldGateValue = [inputGate booleanValue];
	
	return YES;
}

- (NSDictionary*)state
{
	NSMutableDictionary *stateDict = [[NSMutableDictionary alloc] initWithCapacity:3];
	[stateDict addEntriesFromDictionary:[super state]];
	
	[stateDict setObject:[NSNumber numberWithBool:cgMode] forKey:@"CoreGraphicsMode"];
	//NSLog(@"state: %@",stateDict);
	return stateDict;
}

- (BOOL)setState:(NSDictionary*)state
{
	cgMode = [[state objectForKey:@"CoreGraphicsMode"] boolValue];
	return [super setState:state];
}


- (void)toggleCGMode
{
	cgMode ^= 1;
}

- (BOOL)cgMode
{
	return cgMode;
}

@end
